@@ -1,17 +1,22 @@
 import cv2
 import json
 import numpy as np
+import pandas as pd
 import os
 import paho.mqtt.client as mqtt
 import paho.mqtt.subscribe as subscribe
 import queue
 import re
+import sys
 import tarfile
+import threading
 import time
+import requests
+import multiprocessing
 from base64 import b64encode
 from fair_research_login.client import NativeClient
 from mdml_client.config import CLIENT_ID
-from threading import Thread
+
 
 def print_MDML_message(client, userdata, message):
     # self.file_queue.put(message)
@@ -53,7 +58,7 @@ def unix_time(ret_int=False):
     else:
         return unix_time
 
-def read_image(file_name, resize_x=0, resize_y=0):
+def _read_image_file(file_name, resize_x=0, resize_y=0, rescale_pixel_intensity=False):
     """
     Read image from a local file and convert to bytes from sending
     over the MDML.
@@ -88,6 +93,122 @@ def read_image(file_name, resize_x=0, resize_y=0):
     img_byte_string = img_b64bytes.decode('utf-8')
     return img_byte_string
 
+def read_image(file_name, resize_x=0, resize_y=0, rescale_pixel_intensity=False):
+    """
+    Read image from a local file and convert to bytes from sending
+    over the MDML.
+
+
+    Parameters
+    ----------
+    file_name : string
+        file name of the file to be opened
+    resize_x : int
+        horizontal size of the resized image
+    resize_y : int
+        vertical size of the resized image
+    rescale_pixel_intensity : bool
+        scale the pixel intensities between 0-255 according to the min and max pixel values.
+
+    Returns
+    -------
+    string
+        String of bytes that can be used as the second argument in 
+        experiment.publish_image()
+    """
+    source = cv2.imread(file_name, cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH) 
+    if resize_x != 0 and resize_y != 0:
+        source = cv2.resize(source, (resize_x, resize_y))
+    if rescale_pixel_intensity:
+        source = np.nan_to_num(source)
+        min_val = np.min(source)
+        max_val = np.max(source)
+        source = (source - min_val) * (255/(max_val - min_val))
+    _, img = cv2.imencode('.jpg', source)
+    img_b64bytes = b64encode(img)
+    img_byte_string = img_b64bytes.decode('utf-8')
+    return img_byte_string
+
+def GET_images(image_metadata, experiment_id, host):
+    """
+    Retrieves images from the MDML using a GET request. Created for
+    use in FuncX functions that analyze images.
+
+
+    Parameters
+    ----------
+    image_metadata : list
+        list of dicts of image metadata. From the query value for the image-generating device
+    host : string
+        Hostname/IP of the MDML host
+    experiment_id : string
+        MDML Experiment ID
+
+    Returns
+    -------
+    list
+        List of bytes, one element for each image from the metadata
+    """
+    imgs = []
+    for img in image_metadata:
+        resp = requests.get(f"http://{host}:1880/image?path={img['filepath']}&experiment_id={experiment_id}")
+        imgs.append(resp.content)
+    return imgs
+
+def query_to_pandas(device, query_result, sort=True):
+    """
+    Pull out a device's queried data as a pandas DataFrame from the results of experiment.query().
+    
+    Parameters
+    ----------
+    device : string
+        String of the device id that you want to pull out as a pandas DataFrame
+    query_result : list
+        Result of a experiment.query() call.
+
+    Returns
+    -------
+    DataFrame
+        Pandas DataFrame of the query performed
+    """
+    try:
+        device_data = query_result[device]
+        if len(device_data) == 0:
+            raise Exception("No data found in the given device.")
+    except KeyError:
+        print("Device does not exist in this query result.")
+    
+    tmp = {i:[] for i in device_data[0]}
+    for row in device_data:
+        for d in row:
+            tmp[d].append(row[d])
+    device_data_pd = pd.DataFrame.from_dict(tmp)
+    if sort:
+        device_data_pd = device_data_pd.sort_index(axis=1)
+    return device_data_pd
+
+def query(query, experiment_id, host, params={}):
+    """
+    Query the MDML for an example of the data structure that your query will return. This is aimed at aiding in development of FuncX functions for use with the MDML.
+
+    Parameters
+    ----------
+    query : list
+        Description of the data to send funcx. See queries format in the documentation on GitHub
+    experiment_id : string
+        MDML experiment ID for which the data belongs
+    host : string
+        Host of the MDML instance
+    
+    Returns
+    -------
+    list
+        Data structure that will be passed to FuncX
+    """
+    import json
+    resp = requests.get(f"http://{host}:1880/query?query={json.dumps(query)}&parameters={json.dumps(params)}&experiment_id={experiment_id}")
+    return json.loads(resp.text)
+
 class experiment:
     """
     This is the first step in interacting with the MDML. 
@@ -113,7 +234,9 @@ class experiment:
         self.host = host
         self.port = 1883
         self.tokens = None
-        self.msg_queue = queue.Queue()
+        self.msg_queue = None
+        self.debugger = None
+        self.debug_callback = None
         
 
         # Creating connection to MQTT broker
@@ -170,68 +293,69 @@ class experiment:
             with open(config, 'r') as config_file:
                 config_str = config_file.read()
             try:
-                self.config = json.loads(config_str)
+                config = json.loads(config_str)
             except:
                 print("Error in json.loads() call on the config file contents.")
                 return
         elif type(config) == dict:
-            self.config = config
+            config = config
         else:
             print("Supplied configuration type is not supported. Must be of type str or type dict.")
 
-        config_keys = self.config.keys()
-
         # Validating top level section
-        if 'experiment' not in config_keys or 'devices' not in config_keys:
+        if 'experiment' not in config or 'devices' not in config:
             print("""Highest level of configuration json must be a dictionary 
             with the keys: 'experiment' and 'devices'""")
             return False
         
         # Validating experiment section
-        experiment_keys = self.config['experiment']
+        experiment_section = config['experiment']
             # 'experiment_number' not in experiment_keys or\
-        if 'experiment_id' not in experiment_keys or\
-            'experiment_notes' not in experiment_keys or\
-            'experiment_devices' not in experiment_keys:
+        if 'experiment_id' not in experiment_section or\
+            'experiment_notes' not in experiment_section or\
+            'experiment_devices' not in experiment_section:
             print("""Missing required fields in the 'experiment' section of your
             configuration""")
             return False
 
         # Validating devices section
-        devices = self.config['devices']
-        for device_keys in devices:
-            if 'device_id' not in device_keys or\
-                'device_name' not in device_keys or\
-                'device_output' not in device_keys or\
-                'device_output_rate' not in device_keys or\
-                'device_data_type' not in device_keys or\
-                'device_notes' not in device_keys or\
-                'headers' not in device_keys or\
-                'data_types' not in device_keys or\
-                'data_units' not in device_keys:
+        devices = config['devices']
+        for device in devices:
+            if 'device_id' not in device or\
+                'device_name' not in device or\
+                'device_output' not in device or\
+                'device_output_rate' not in device or\
+                'device_data_type' not in device or\
+                'device_notes' not in device or\
+                'headers' not in device or\
+                'data_types' not in device or\
+                'data_units' not in device:
                 print("""Missing required fields in the 'devices' section of your 
                 configuration""")
                 return False
 
         if self.tokens:
             try:
-                self.config['globus_token'] = self.tokens['funcx_service']['access_token']
+                config['globus_token'] = self.tokens['funcx_service']['access_token']
             except:
                 print("No Auth token found. Have you run .globus_login() to create one?\
                     This can be ignored if you are not using funcX for analysis.")
                 pass
 
         if experiment_run_id != "":
-            self.config['experiment']['experiment_run_id'] = experiment_run_id
-        # Check run id only contains letters and underscores
-        if re.match(r"^[\w]*$", self.config['experiment']['experiment_run_id']):
-            self.config = json.dumps(self.config)
-            print("Valid configuration found, now use .send_config() to send it to the MDML.")
-            # Return to string to prepare for sending to MDML
-            return True
-        else:
-            print("Experiment run ID contains characters other than letters, numbers, and underscores.")
-            return False
+            # Check run id only contains letters and underscores
+            if re.match(r"^[\w]*$", experiment_run_id):
+                config['experiment']['experiment_run_id'] = experiment_run_id
+                # Set config as string to prepare for sending to MDML
+                self.config = json.dumps(config)
+                print("Valid configuration found, now use .send_config() to send it to the MDML.")
+                return True
+            else:
+                print("Experiment run ID contains characters other than letters, numbers, and underscores.")
+                return False
+        else: # No need to add a blank run ID
+            # Set config as string to prepare for sending to MDML
+            self.config = json.dumps(config) 
 
     def send_config(self):
         """
@@ -298,7 +422,7 @@ class experiment:
         # Send data via MQTT
         self.client.publish(topic, json.dumps(payload))
       
-    def publish_data(self, device_id, data, data_delimiter='null', influxDB = True):
+    def publish_data(self, device_id, data, data_delimiter='null', influxDB = True, timestamp = 0):
         """
         Publish data to MDML
         
@@ -328,33 +452,39 @@ class experiment:
             payload['data_delimiter'] = data_delimiter
         if influxDB:
             payload['influx_measurement'] = device_id.upper()
-        
+        # Added for system timings
+        payload['timestamp'] = unix_time()
         # Send data via MQTT
         self.client.publish(topic, json.dumps(payload))
 
-    def publish_analysis(self, queries, function_id, endpoint_id):
+    def publish_analysis(self, device_id, queries, function_id, endpoint_id, parameters={}):
         """
         Publish a message to run an analysis
 
 
         Parameters
         ----------
+        device_id : string
+            Device ID for storing analysis results (must match configuration file)
         queries : list
             Description of the data to send funcx. See queries format in the documentation on GitHub
         function_id : string
             From FuncX, the id of the function to run
         endpoint_id : string
             From FuncX, the id of the endpoint to run the function on
+        parameters : any json serializable type
+            Custom parameters to be accessed in the second element of your FuncX data parameter
         """
         # Creating MQTT topic
-        topic = "MDML/" + self.experiment_id + "/FUNCX"
+        topic = "MDML/" + self.experiment_id + "/FUNCX/" + device_id
 
         # Set message payload
         payload = {
             'queries': queries,
             'function_id': function_id,
             'endpoint_id': endpoint_id,
-            'timestamp': unix_time()
+            'timestamp': unix_time(),
+            'parameters': parameters
         }
 
         # Add auth if set
@@ -368,7 +498,81 @@ class experiment:
         # Send data via MQTT
         self.client.publish(topic, json.dumps(payload))
         
-    def publish_image(self, device_id, img_byte_string, filename = '', timestamp = 0):
+    def publish_image(self, device_id, img_byte_string, filename = '', timestamp = 0, metadata = {}):
+        """
+        Publish an image to MDML
+
+
+        Parameters
+        ----------
+        device_id : str
+            Unique string identifying the device that created the data.
+            This must correspond with the experiment's configuration
+        img_byte_string : str
+            byte string of the image you want to send. Can be supplied by
+            mdml_client.read_image() function in this package
+        filename : str
+            filename to store the file in the MDML. Can only contain letters, 
+            numbers, underscores and must end with a valid file extension. 
+            If left blank, filenames will the experiment ID followed by an
+            index (e.g. EXPID_1.JPG, EXPID_2.JPG...)
+        timestamp : int
+            Unix time in nanoseconds. Can be supplied by the unix_time()
+            function in this package
+        metadata : dict
+            Dictionary containing any metadata for the image. Data types of 
+            the dictionary values must not be changed. Keys cannot include
+            "time" or "filepath".
+        """
+
+        # Creating MQTT topic
+        topic = "MDML/" + self.experiment_id + "/DATA/" + device_id.upper()
+        # Base payload
+        payload = {
+            'filename': filename,
+            'data': img_byte_string,
+            'data_type': 'image'
+        }
+        # Adding metadata if necessary
+        payload['metadata'] = metadata
+        # Check for valid filename
+        if filename != '':
+            if re.match(r"^[\w]+\.[A-Za-z0-9]+$", filename) == None:
+                print("Filename not valid. Can only contains letters, numbers, and underscores.")
+                return
+            else:
+                payload['filename'] = filename
+        # Adding timing code
+        if timestamp == 0:
+            timestamp = unix_time()
+        payload['timestamp'] = timestamp
+        payload = json.dumps(payload)
+        # Publish it
+        self.client.publish(topic, payload)
+
+    def query(query, host, params={}):
+        """
+        Query the MDML for an example of the data structure that your query will return. This is aimed at aiding in development of FuncX functions for use with the MDML.
+
+        Parameters
+        ----------
+        query : list
+            Description of the data to send funcx. See queries format in the documentation on GitHub
+        experiment_id : string
+            MDML experiment ID for which the data belongs
+        host : string
+            Host of the MDML instance
+        
+        Returns
+        -------
+        list
+            Data structure that will be passed to FuncX
+        """
+        import json
+        resp = requests.get(f"http://{host}:1880/query?query={json.dumps(query)}&parameters={json.dumps(params)}&experiment_id={self.experiment_id}")
+        return json.loads(resp.text)
+
+    def _publish_image_benchmarks(self, device_id, img_byte_string, filename = '', timestamp = 0, size = 0):
         """
         Publish an image to MDML
 
@@ -388,6 +592,8 @@ class experiment:
         timestamp : int
             Unix time in nanoseconds. Can be supplied by the unix_time()
             function in this package
+        size : string
+            Size of the message being sent
         """
 
         # Creating MQTT topic
@@ -400,7 +606,8 @@ class experiment:
             'timestamp': timestamp,
             'filename': filename,
             'data': img_byte_string,
-            'data_type': 'image'
+            'data_type': 'image',
+            'size': size
         }
         # Check for valid filename
         if filename != '':
@@ -426,7 +633,7 @@ class experiment:
         Init an MDML debugger to retrieve error messages or other important 
         events when running an experiment.
         """
-        debug = Thread(target=subscribe.callback,\
+        self.debugger = multiprocessing.Process(target=subscribe.callback,\
             kwargs={\
                 'callback': on_MDML_message,\
                 'topics': "MDML_DEBUG/" + self.experiment_id,\
@@ -434,8 +641,14 @@ class experiment:
                 'auth': {'username': self.username, 'password': self.password},\
                 'userdata': self.msg_queue
             })
-        debug.setDaemon(False)
-        debug.start()
+        self.debugger.start()
+
+    def disconnect(self):
+        """
+        Disconnect MQTT client from the broker
+        """
+        self.client.disconnect()
+        print("Disconnected from MDML.")
 
     def set_debug_callback(self, user_func):
         """
@@ -449,13 +662,25 @@ class experiment:
             the MDML will trigger this function with the message string as 
             the parameter.
         """
+        self.msg_queue = multiprocessing.Manager().Queue()
         def func(msg_queue):
             while True:
                 msg = msg_queue.get()
                 user_func(msg)
-        debug_callback = Thread(target=func, args=(self.msg_queue,))
-        debug_callback.setDaemon(False)
-        debug_callback.start()
+                sys.stdout.flush()
+        self.debug_callback = multiprocessing.Process(target=func, args=(self.msg_queue,))
+        self.debug_callback.start()
+
+    def stop_debugger(self):
+        """
+        Stop an MDML debugger that has already been started
+        """
+        print("Stopping debugger.")
+        if self.debug_callback.is_alive():
+            self.debug_callback.terminate()
+        if self.debugger.is_alive():
+            self.debugger.terminate()
+        print("Debugger stopped.")
 
     def replay_experiment(self, filename):
         """
@@ -515,7 +740,7 @@ class experiment:
         try:
             # Start simulations
             for i in range(len(valid_devices)):
-                tmp = Thread(target=self._replay_file, args=(valid_devices[i], \
+                tmp = threading.Thread(target=self._replay_file, args=(valid_devices[i], \
                                                         exp_dir, \
                                                         device_data_types[i], \
                                                         exp_start_time, \
@@ -554,4 +779,4 @@ class experiment:
                     print("DATA_TYPE IN CONFIGURATION NOT SUPPORTED")
                 del data[0]
                 if len(data) == 0:
-                    break
\ No newline at end of file
+                    break
